<?php
/**
 * SpecialHierarchy.php
 *
 * This file is part of the Hierarchy extension for MediaWiki; supporting the
 * Treeview skin.  Its code performs two main functions: ensuring cache-
 * consistency and providing server-side support for client-side load-on-demand
 * of treeview branches.
 *
 * Licenced under the General Public Licence 2, without warranty.
 *
 * @licence GPL2 http://www.gnu.org/copyleft/gpl.html
 * @author http://members.dodo.com.au/~netocrat
 * @version $Date: 2007/03/22 08:36:51 $ $Revision: 1.19 $ (development/testing)
 */
define('TV_RELEASE_VERSION', '1.1.beta');
# Register extension
$wgExtensionFunctions[] = 'wfSpecialHierarchy';
$wgExtensionCredits['specialpage'][] = array(
	'name' => 'Hierarchy',
	'author' => '[http://members.dodo.com.au/~netocrat Netocrat]',
	'version' => TV_RELEASE_VERSION,
	'url' => 'http://clc-wiki.net/wiki/Planning:Treeview_skin'
);

global $IP; # not required for MW 1.5.x or 1.6.x
require_once('Hierarchy.php');
/** for FauxRequestEx */
require_once('State.php');
require_once("$IP/includes/SpecialPage.php");
/** for wfGetRedirectsTo() and wfGetSeeAlsos() */
require_once('UtilityFunctions.php');
/** for TvQueryString::queryStrToArr() */
require_once('QueryString.php');

# MediaWiki's devoid of a hook for undelete and for before/after category link
# updates so patches are necessary there; earlier versions (<1.6?) are also
# devoid of a hook for file uploads.  Patches are supplied in Hierarchy/patches.
$wgHooks['ArticleDeleteComplete'][] = 'wfHierarchy_ArticleDeleteCompleteHook';
$wgHooks['UndeleteComplete'][] = 'wfHierarchy_UndeleteCompleteHook';
$wgHooks['ArticleSave'][] = 'wfHierarchy_ArticleSaveHook';
$wgHooks['ArticleSaveComplete'][] = 'wfHierarchy_ArticleSaveCompleteHook';
$wgHooks['TitleMoveComplete'][] = 'wfHierarchy_TitleMoveCompleteHook';
$wgHooks['UploadComplete'][] = 'wfHierarchy_UploadCompleteHook';
$wgHooks['BeforeCategoryLinksUpdate'][]='wfHierarchy_BeforeCategoryLinksUpdate';
$wgHooks['AfterCategoryLinksUpdate'][] = 'wfHierarchy_AfterCategoryLinksUpdate';
$wgHooks['UnknownAction'][] = 'wfHierarchy_UnknownActionHook';
$wgHooks['SkinTemplateTabs'][] = 'wfHierarchy_SkinTemplateTabsHook';

/** Adds tabs for the rebuildtreeview, parsetreeview, treeviewcachestatus
  * actions when the current page is the (pseudo-) Treeview/Hierarchy
  * configuration page.
  */
function wfHierarchy_SkinTemplateTabsHook(&$skin, &$content_actions) {
	global $wgTitle, $wgRequest, $wgUser;
	$action = $wgRequest->getText('action');
	$res = Hierarchy::isConfigPage($wgTitle);
	if ($res !== false) {
		if ($wgUser->isAllowed('parsetreeview')) {
			$content_actions['parsetreeview'] = array(
				'class' => $action == 'parsetreeview' ?
				            'selected' : 'tvaction',
				'text'  => wfMsgHTML('tv_parsetreeview'),
				'href'  => $wgTitle->getLocalUrl(
				            'action=parsetreeview')
			);
		}
		if ($wgUser->isAllowed('rebuildtreeview')) {
			$content_actions['rebuildtreeview'] = array(
				'class' => $action == 'rebuildtreeview' ?
				            'selected' : 'tvaction',
				'text'  => wfMsgHTML('tv_rebuildtreeview'),
				'href'  => $wgTitle->getLocalUrl(
				            'action=rebuildtreeview')
			);
		}
		if ($wgUser->isAllowed('treeviewcachestatus')) {
			$content_actions['treeviewcachestatus'] = array(
				'class' => $action == 'treeviewcachestatus' ?
				            'selected' : 'tvaction',
				'text'  => wfMsgHTML('tv_treeviewcachestatus'),
				'href'  => $wgTitle->getLocalUrl(
				            'action=treeviewcachestatus')
			);
		}
	}
	return true;
}

/** This can't be hooked in the SkinTreeview file since that file only gets
  * included at a later point than this function needs to be called.
  */
$wgHooks['ArticleViewRedirect'][] ='wfTreeviewSkin_ArticleViewRedirect';
function wfTreeviewSkin_ArticleViewRedirect($article) {
	global $wgUser;
	$sk = $wgUser->getSkin();
	return $sk->mDisableDefaultRedirectSubHeadings ? false : true;
}
/** Implements the rebuildtreeview, parsetreeview, treeviewcachestatus actions
  * as reached through the (pseudo-) configuration page's corresponding tabs:
  * 'Parse configuration', 'Rebuild treeview', 'Cache status'
  */
function wfHierarchy_UnknownActionHook($action, $article) {
	global $wgOut, $wgUser, $wgHierarchyPage;
	if ($action == 'view_tv') {
		global $wgUseFileCache;
		wfDebug('wfHierarchy_UnknownActionHook: setting '.
		  "$wgUseFileCache false.\n");
		$wgUseFileCache = false;
		$article->view();
		$handled = true;
	} else if ($action == 'rebuildtreeview' || $action == 'parsetreeview' ||
	  $action == 'treeviewcachestatus') {
		$handled = true;
		if (!$wgUser->isAllowed($action)) {
			$handled = false;
			$wgOut->permissionRequired($action);
		} else {
			$res = Hierarchy::isConfigPage($article->getTitle());
			if ($res === false) {
				$sk = $wgUser->getSkin();
				/* This repeats the title creation within
				 * isConfigPage() but it is rarely reached and
				 * is not time-critical.
				 */
				$tConfig = Hierarchy::getConfigPageTitle(true);
				$configHref = $tConfig ?
				  $sk->makeLinkObj($tConfig) :
				  wfMsg('tv_invalid').wfMsg('tv_invconfpg',
				    $wgHierarchyPage);
				$wgOut->addHTML(wfMsg('tv_err_actionforconf',
				  $action, $configHref));
			} else {
				$s = '<h1>'.wfMsg('tv_'.$action)."</h1>\n";
				$hObj = new Hierarchy;
				if ($action == 'rebuildtreeview' ||
				    $action == 'treeviewcachestatus') {
					/* This is called prior to the call to
					 * parseConfigPage() so that a re-parse
					 * is avoided. */
					$cacheHit = $hObj->loadHierarchy();
					$ts = $hObj->getHierarchyTs();
				}
				if ($action != 'treeviewcachestatus') {
					$parseRes = $hObj->parseConfigPage();
				}
				if ($res === 0) {
					$s .= '<h2>'.wfMsg('tv_standin_hdr').
					  "</h2>\n".
					  wfMsg('tv_standinconfpg_desc')."\n";
				}
				if ($action == 'rebuildtreeview') {
					/* The parse result is stored in the
					 * object so that this call does not
					 * invoke a re-parse. */
					$hObj->buildHierarchyAndRemappings();
					$hObj->saveRemappings();
					$hObj->saveHierarchy();
					$cacheHit2 = $hObj->loadHierarchy();
					$ts2 = $hObj->getHierarchyTs();
					$k = $cacheHit2 ?
					  'tv_rebuildsucc' : 'tv_rebuildmiss';
					$postStat = '<h2>'.wfMsg($k.'_hdr').
					  "</h2>\n".wfMsg($k).'<h2>'.wfMsg(
					  'tv_cachestatpost_hdr')."</h2>\n".
					  wfMsg($cacheHit2 === 0 ?
					   'tv_unusablecachehit' : ($cacheHit2 ?
					    'tv_cachehit':'tv_cachemiss'),$ts2).
					  "\n";
				} else {
					$cacheHit2 = false;
					$postStat = '';
				}
				/* Output the result of the cache retrieval(s).
				 * Warn if the cache seems non-functional. */
				if ($action == 'rebuildtreeview' ||
				    $action == 'treeviewcachestatus') {
					global $wgMainCacheType;
					if ($cacheHit2 === false &&
					  $cacheHit === false &&
					  $wgMainCacheType == CACHE_NONE) {
						$s .= '<h2>'.wfMsg(
						  'tv_nocache_hdr')."</h2>\n".
						  wfMsg('tv_nocache');
					}
					$s .= $postStat;
					$hdrKey = $action == 'rebuildtreeview' ?
					  'tv_cachestatpre_hdr' :
					  'tv_cachestatus_hdr';
					$s .= '<h2>'.wfMsg($hdrKey)."</h2>\n".
					  wfMsg($cacheHit === 0 ?
					    'tv_unusablecachehit' : ($cacheHit ?
					     'tv_cachehit':'tv_cachemiss'),$ts).
					  "\n";
				}
				/* Output parse diagnostics. */
				if ($action != 'treeviewcachestatus') {
					$s .= '<h2>'.wfMsg('tv_parseerrintro').
					  "</h2>\n".($parseRes ?
					    wfMsg('tv_noproblems') :
					    $hObj->getParseErrors_Html());
				}
				$s .= '<h1>'.wfMsg('tv_content')."</h1>\n";
				$wgOut->addHTML("$s\n");
				$wgOut->addWikiText($hObj->getConfigPageText());
			}
		}
	} else	$handled = false;
	/* For MediaWiki 1.6.5, a return value of false is interpreted as
	 * 'unhandled'; for MediaWiki 1.5.6 and 1.7.1, a return value of true
	 * is interpreted in that way.  Returning nothing avoids this
	 * contradiction, although it means that for unhandled actions a blank
	 * page is displayed instead of the 'unrecognised action' message.
	 */
//	return !$handled;
}

function wfHierarchy_ArticleDeleteCompleteHook() {
	wfDebug('wfHierarchy_ArticleDeleteCompleteHook: Invoking '.
	  "indiscriminate invalidation.\n");
	HierarchyPage::invalidateAll();
}

function wfHierarchy_UndeleteCompleteHook() {
	wfDebug('wfHierarchy_UndeleteCompleteHook: Invoking indiscriminate '.
	  "invalidation.\n");
	HierarchyPage::invalidateAll();
}

function wfHierarchy_TitleMoveCompleteHook() {
	wfDebug('wfHierarchy_TitleMoveCompleteHook: Invoking indiscriminate '.
	  "invalidation.\n");
	HierarchyPage::invalidateAll();
}

/**
 * The one and only non-configuration global used by Treeview-related code: it's
 * used only to work around the parts of MediaWiki's hook system that aren't
 * designed for Treeview purposes.
 */
$wgSpecialHierarchySavedHookData = array('invalidatedall' => false);

function wfHierarchy_ArticleSaveHook($article) {
	global $wgSpecialHierarchySavedHookData;
	$fname = 'wfHierarchy_ArticleSaveHook';
	$t = $article->getTitle();
	$id = $t->getArticleID();
	# Trigger on all page creations (this test could be refined) ...
	$invalidate = ($id == 0);
	if ($invalidate) wfDebug("$fname: detected a page creation\n");
	else {
		# ...and on edits to the hierarchy config page...
		if (Hierarchy::isConfigPage($t)) {
			$invalidate = true;
			wfDebug("$fname: detected a save to the config page\n");
		}
	}
	# ... and on edits to any messages upon which that config depends
	if (!$invalidate) {
		foreach (Hierarchy::getMessageDeps() as $msg) {
			# MediaWiki always capitalises the first letter of
			# articles in the message namespace upon save even when
			# $wgCapitalLinks = false
			$tMsg = Title::newFromText(ucfirst($msg), NS_MEDIAWIKI);
			if ($id == $tMsg->getArticleID()) {
				$invalidate = true;
				wfDebug("$fname: detected a saved edit to the ".
				  "message '$msg', upon which the config page ".
				  "depends\n");
				break;
			}
		}
	}
	if ($invalidate) {
		HierarchyPage::invalidateAll();
		$hookData['invalidatedall'] = true;
	} else {
		# ... and possibly later if this is a changing redirect
		$rev = Revision::newFromTitle($t);
		$oldText = $rev->getText();
		if (($tTarget = Title::newFromRedirect($oldText))) {
			$hookData['old_target'] = $tTarget;
			$a = wfGetRedirectsTo($tTarget);
			$a[] = $tTarget;
			$hookData['old_aliaslist'] = $a;
		}
	}
	$wgSpecialHierarchySavedHookData = $hookData;
	return true;
}

function wfHierarchy_ArticleSaveCompleteHook($article, $user, $text) {
	global $wgSpecialHierarchySavedHookData;
	$fname = 'wfHierarchy_ArticleSaveCompleteHook';
	$hookData = $wgSpecialHierarchySavedHookData;
	if ($hookData['invalidatedall']) return true;
	$t = $article->getTitle();
	$tTarget = Title::newFromRedirect($text);
	$tOldTarget = $hookData['old_target'];
	# If this was a redirect that post-save no longer redirects to the same
	# page ...
	if ($tOldTarget && (!$tTarget ||
	  $tOldTarget->getArticleID() != $tTarget->getArticleID())) {
		# ...then invalidate the article that it previously redirected
		# to and all other articles that redirected to that article.
		$invalArr = $hookData['old_aliaslist'];
		$titleInList = true;
	} else{
		$invalArr = array();
		$titleInList = false;
	}
	# If this is now a redirect that points to a different page than
	# pre-save ...
	if ($tTarget && (!$tOldTarget ||
	  $tOldTarget->getArticleID() != $tTarget->getArticleID())) {
		# ...then invalidate the article that it now redirects to and
		# all other articles that redirect to that article.
		$a = wfGetRedirectsTo($tTarget);
		if ($titleInList) $a[] = $tTarget;
		$invalArr = array_merge($invalArr, $a);
	}
	if ($invalArr) {
		$dbgList = '';
		foreach ($invalArr as $t) {
			if ($dbgList) $dbgList .= ', ';
			$dbgList .= $t->getPrefixedText();
		}
		wfDebug("$fname: detected a new or changed redirect - ".
		  "INVALIDATING the cache of these articles: $dbgList\n");
		/** @note Ideally this wouldn't affect the parser cache. */
		Title::touchArray($invalArr);
	}
	return true; # warning: prior conditional exit
}

function wfHierarchy_BeforeCategoryLinksUpdate($nt, $old, $new, &$oSeeAlsos) {
	global $wgSpecialHierarchySavedHookData;
	if ($wgSpecialHierarchySavedHookData['invalidatedall']) return true;
	# MW 1.5.x and 1.6.x differ; this tests for the earlier behaviour
	if (!is_object($nt)) $nt = Title::newFromText($nt);
	$oSeeAlsos = $nt ? wfGetSeeAlsos($nt, true) : array();
	return true;
}

function wfHierarchy_AfterCategoryLinksUpdate($nt, $old, $new, $oSeeAlsos) {
	global $wgSpecialHierarchySavedHookData;
	if ($wgSpecialHierarchySavedHookData['invalidatedall']) return true;
	$fname = 'wfHierarchy_AfterCategoryLinksUpdate';
	# MW 1.5.x and 1.6.x differ; this tests for the earlier behaviour
	if (!is_object($nt)) $nt = Title::newFromText($nt);
	$nSeeAlsos = $nt ? wfGetSeeAlsos($nt, true) : array();
	$removed = array_diff_key($oSeeAlsos, $nSeeAlsos);
	$added = array_diff_key($nSeeAlsos, $oSeeAlsos);
	$changedList = $removed + $added;
	/* In later versions of MW these arrays aren't old and new but rather
	 * deleted and added categories. */
	if ($old || $new) HierarchyPage::invalidateAll();
	else if ($changedList) {
		$invalArr = array();
		foreach ($changedList as $page_id => $pg) {
			$t = Title::makeTitle($pg['page_namespace'],
			  $pg['page_title']);
			if ($t) $invalArr[] = $t;
		}
		if ($invalArr) {
			$dbgList = '';
			foreach ($invalArr as $t) {
				if ($dbgList) $dbgList .= ', ';
				$dbgList .= $t->getPrefixedText();
			}
			wfDebug("$fname: INVALIDATING the cache of these ".
			  "articles due to see-also list changes: $dbgList\n");
			/** @note As above - ideally, ignore the parser cache.*/
			Title::touchArray($invalArr);
		}
	}
}

/**
 * This is overkill when the image already exists - as long as its url doesn't
 * change there's no need to rebuild things - but if the upload is for a new
 * image then this invalidation is required because the url will have been
 * unknown to the treeview.  An optimisation would be to distinguish between
 * the two.
 */
function wfHierarchy_UploadCompleteHook($img) {
	$fname = 'wfHierarchy_UploadCompleteHook';
	$imageName = $img->getName();
	if (Hierarchy::usesImage($imageName)) {
		wfDebug("$fname: Uploaded image '$imageName' is used by the ".
		  "treeview: invoking indiscriminate invalidation.\n");
		HierarchyPage::invalidateAll();
	}
}

/** This class implements a special page used by the Treeview skin to download
  * through javascript only the necessary nodes when expanding an unloaded part
  * of the treeview.
  * @package MediaWiki
  * @subpackage Extensions
  */
class HierarchyPage extends SpecialPage {
	/**#@+ @access public */

	/** Constructor for HierarchyPage. */
	function HierarchyPage() {
		SpecialPage::SpecialPage('Hierarchy');
	}

	/** Main entry point function for HierarchyPage as a SpecialPage. */
	function execute($par) {
		$fname = 'HierarchyPage::execute';

		/* Avoid $wgRequest because it includes cookies. */
		$queryParams = $_GET;
		unset($queryParams['title']);
		if (!$queryParams) {
			$this->showMainControls();
		} else	$this->handleLoadOnDemand_Xml();
	}

	function showMainControls() {
		global $wgOut, $wgUser, $wgHierarchyPage;
		$this->setHeaders();
		$wgOut->addHTML('<h1>Hierarchy/Treeview controls</h1>'."\n");
		list($tConfig, $fullMatch) = Hierarchy::getConfigPageTitle_();
		if (!$tConfig) {
			$wgOut->addHTML(wfMsg('tv_invconfpg', $wgHierarchyPage).
			  "\n");
		} else {
			$sk = $wgUser->getSkin();
			$confPage = $fullMatch? 'confpg': 'standinconfpg';
			$confTitle = $tConfig->getPrefixedText();
			$wgOut->addHTML('<dl>'."\n");
			foreach (array(
			  $confPage             => false,
			  'parsetreeview'       => true,
			  'rebuildtreeview'     => true,
			  'treeviewcachestatus' => true)
			as $key => $isAction) {
				if ($isAction && !$wgUser->isAllowed($key)) {
					continue;
				}
				$desc = wfMsg('tv_'.$key.'_desc');
				$linkFunc = $isAction ?
				  'makeKnownLinkObj' : 'makeLinkObj';
				$link = $sk->$linkFunc($tConfig,
				  wfMsg('tv_'.$key, $confTitle),
				  $isAction?"action=$key":'');
				$wgOut->addHTML("<dt>$link</dt>\n".
				  "<dd>$desc</dd>\n");
			}
			$wgOut->addHTML('<dl>'."\n");
		}
	}

	/** Handles a load-on-demand request for a (set of) nodes.  Interprets
	  * the url query parameters, builds the branch(es) and outputs XML in
	  * place of MediaWiki's default output mechanisms.  The request is
	  * issued, and its result interpreted, by code in
	  * [mediawikiroot]/skins/treeview/treeview.js
	  */
	function handleLoadOnDemand_Xml() {
		global $wgOut, $wgInputEncoding, $wgRequest, $wgUser;
		$fname = 'HierarchyPage::handleLoadOnDemandXml';
		/* Avoid E_NOTICE warnings during accesses to undefined array
		 * keys.*/
		wfSuppressWarnings();

		$nodeIds = explode(',', $wgRequest->getText('expnodeids'));
		$base = $wgRequest->getVal('base');
		/* WebRequest::normalizeUnicode performs (all/most of?) the
		 * equivalent of urldecode. */
		$effFullUrl = /*urldecode(*/$wgRequest->getVal('effurl')/*)*/;
		$tmp = strpos($effFullUrl, '?');
		$effQueryStr = $tmp!==false ? substr($effFullUrl, $tmp+1) : '';
		$fauxParams = TvQueryString::queryStrToArr($effQueryStr, false);
		if ($base && preg_match("#$base/(.*)#",$effFullUrl,$effTitle)) {
			$effTitle = $effTitle[1];
		} else {
			$effTitle = $wgRequest->getVal('effttl');
			/* When a url does not contain a title, e.g.
			 * index.php?action=edit, then MediaWiki uses the Main
			 * Page title. */
			if (!$effTitle) $effTitle = $fauxParams['title'] ?
			  $fauxParams['title'] : wfMsgNoDB('mainpage');
		}
		$tEffective = Title::newFromText($effTitle);
		$effWebRequest = new FauxRequestEx($fauxParams, $effFullUrl);
		$state = new TVstate(array(
			'tEffective' => $tEffective,
			'tRedirectInfo' => null, # unknown: follow on-demand
			'effectiveWebRequest' => $effWebRequest,
			'effectiveQueryStr' => $effQueryStr,
			'effectiveAction' => $effWebRequest->getVal('action')));
		$hierarchy = new Hierarchy($state);
		$skipStateRebuild = $wgRequest->getBool('skiplinkstaterebuild');
		wfDebug("$fname: retrieved from request :-\n".
		  "*effective title     : ".($tEffective
		    ? ('"'.$tEffective->getPrefixedText().'"')
		    : '[could not build Title from "'.$effTitle.'"]')."\n".
		  "*base                : ".($base ? "\"$base\"" : 'n/a')."\n".
		  "*expand-node ids     : ".var_export($nodeIds, true)."\n".
		  "*effective full url  : \"$effFullUrl\"\n".
		  "*effective query str : \"$effQueryStr\"\n".
		  '*skiplinkstaterebuild: "'.($skipStateRebuild?'1':'0')."\"\n".
		  "*decoded effective query string as:\n".
		    var_export($fauxParams, true)."\n"
		);

		$op = "<div>\n";
		$htmlArr = $hierarchy->buildHierarchyHtml($nodeIds,
		  $skipStateRebuild, '', true);
		foreach ($htmlArr as $nodeId => $html) {
			if ($html === false) $html = '<input type="hidden" '.
			  "class=\"notfound\">$nodeId</input>";
			$op .= "<div id=\"$nodeId\">$html</div>\n";
		}
		$expandeds = $hierarchy->getExpandedNodesStr();
		$op .= "<form action=\"#\">\n<input type=\"hidden\" ".
		  "id=\"expandednodes\" value=\"$expandeds\" />\n</form>\n";
		$op .= "</div>\n";

		$wgOut->disable();
		header('Content-type: text/xml; charset='.$wgInputEncoding);
//		wfDebug("$fname: returning as xml:\n$op\n");
		echo $op;

		wfRestoreWarnings();
	}

	/**
	 * Invalidates every treeview-related cache and all articles in the
	 * file cache and client-caches; avoids the parser cache.
	 */
	function invalidateAll() {
		global $wgMemc;
		$fname = 'HierarchyPage::invalidateAll';
		wfDebug("$fname: INVALIDATING the cache of all articles + the ".
		  'treeview hierarchy + the remappings list'."\n");
		HierarchyPage::invalidateAll_File_Client();
		$key = Hierarchy::getHierarchyKey();
		$wgMemc->delete($key);
		$key = Hierarchy::getRemappingsKey();
		$wgMemc->delete($key);
	}

	/**
	 * Invalidates all articles in the file cache and client-caches;
	 * avoids the parser cache.
	 */
	function invalidateAll_File_Client() {
		global $wgCacheEpochTouchFile, $wgUseFileCache, $wgCachePages;
		$fname = 'invalidateAll_File_Client';
		if (!@touch($wgCacheEpochTouchFile)) {
			if ($wgUseFileCache || $wgCachePages) {
				global $wgOut;
				$wgOut->fatalError(wfMsg('tv_untouchable',
				  $fname));
			}
		}
	}

	/**#@- */ # end public section
}

/** Encapsulate the hierarchy extension for use as a special page. */
function wfSpecialHierarchy()
{
	global $wgCacheEpoch, $wgCacheEpochTouchFile, $wgUseFileCache,
	  $wgCachePages, $wgLanguageCode, $wgContLanguageCode;
	$fname = 'wfSpecialHierarchy';

	# Default messages in english
	require_once('languages/en.php');

	# Copied and adjusted from elsewhere in MW code.
	wfSuppressWarnings();
	$filename = 'languages/'.
	  str_replace('-', '_', strtolower($wgLanguageCode)).'.php';
	include($filename);
	/* Include this subsequent to the $wgLanguageCode file because the
	 * treeview needs to be generated with a consistent language when
	 * treeview caching is enabled (not tested for here): MW < 1.7.x doesn't
	 * support the second parameter (language code) of
	 * MessageCache::addMessages(), so for those MW versions these
	 * translations need to override the translations set by the previous
	 * include.  A minor side effect is that for those MW versions,
	 * per-user translations of other UI elements are not possible - this is
	 * documented in README.Treeview as #2 under "Limitations".
	 */
	if ($wgLanguageCode != $wgContLanguageCode) {
		$filename = 'languages/'.
		  str_replace('-', '_', strtolower($wgContLanguageCode)).'.php';
		include($filename);
	}
	wfRestoreWarnings();

	$tmp = @filemtime($wgCacheEpochTouchFile);
	if ($tmp !== false) {
		$lastTouched = gmdate('YmdHis', $tmp);
		$org = $wgCacheEpoch;
		$wgCacheEpoch = max($wgCacheEpoch, $lastTouched);
		if ($wgCacheEpoch != $org) {
			wfDebug("$fname: SET \$wgCacheEpoch to $wgCacheEpoch; ".
			  "was $org\n");
		}
	} else if (($wgUseFileCache || $wgCachePages) &&
	  !@touch($wgCacheEpochTouchFile)) {
		global $wgOut;
		$wgOut->fatalError(wfMsg('tv_inaccessibletouch', $fname,
		  '<code>'.$wgCacheEpochTouchFile.'</code>'));
	}

	SpecialPage::addPage(new HierarchyPage);
}

?>
